#include <stdio.h>
#include <stdlib.h>
#include "graphe_p.H"

#define E_INDICE_NOEUD 1
#define E_N 2
#define E_MEMOIRE 3

class arbre_p
{
	private:
	int n;
	int root;
	int *pt;
	int *poidt;
//	Sens des arcs : vrai s'ils s'loignent de la racine.
	char *senst;
	int *dt;
	int *st;

	public:
	arbre_p(int _n,int *_noeuds);
	arbre_p(int _n,int r,graphe_p *graph,arbre_p *arbre_ini);

	int taille();
	int racine();
	int p(int k);
	int poid(int k);
	char sens(int k);
	int d(int k);
	int s(int k);

	void entre(int ex,int ey,int *sx,int *sy,int *min);

	private:
	void copie(int a,int *pt,int *poidt,char *senst,int *dt,arbre_p *arbre_ini);
	char arete(int a,int b,graphe_p *graph);
	void cycle(int ex,int ey,int *am,char *cotex,int *min,int *join);
	void poid_cycle(int ex,int ey,int min,int join);
	int maj_s(int e1,int e2,int f1,int f2);
	void maj_p_poid_sens(int e1,int e2,int min,int se2,int f2);
	void maj_d(int e2,int dernier);
	void maj_arbre(int ex,int ey,int cotex,int min,int am);

	int arbre_p :: rech_dernier(int dsr,int noeud);
	
	public:
	int dernier(int sous_racine);
	
	private:
	void erreur(int i);
};

arbre_p :: arbre_p(int _n,int *noeuds)
{
	int k;
	
	if (_n < 1)
		erreur(E_N);
	
	n = _n;
	root = 1;

	if (!((pt = (int *)calloc(n + 1,sizeof(int))) && 
	(poidt = (int *)calloc(n + 1,sizeof(int))) &&
	(senst = (char *)calloc(n + 1,sizeof(char))) &&
	(dt = (int *)calloc(n + 1,sizeof(int))) &&
	(st = (int *)calloc(n + 1,sizeof(int)))))
		erreur(E_MEMOIRE);
	
	dt[1] = 0;
	st[1] = 2;
	for(k = 2;k < n;k++)
	{
		pt[k] = 1;
		poidt[k] = abs(noeuds[k]);
		if (noeuds[k] < 0)
			senst[k] = 0;
		else
			senst[k] = 1;
		dt[k] = 1;
		st[k] = k + 1;
	}
	pt[n] = 1;
	poidt[n] = abs(noeuds[n]);
	if (noeuds[n] < 0)
		senst[n] = 0;
	else
		senst[n] = 1;
	dt[n] = 1;
	st[n] = 1;
}

void arbre_p :: copie(int a,int *pt,int *poidt,char *senst,int *dt,arbre_p *arbre_ini)
{
	if(a != root)
	{
		pt[a] = arbre_ini->p(a);
		poidt[a] = arbre_ini->poid(a);
		senst[a] = arbre_ini->sens(a);
	}	
	dt[a] = arbre_ini->d(a) - arbre_ini->d(root);
}

char arbre_p :: arete(int a,int b,graphe_p *graph)
{
	return(graph->existe(a,b) || graph->existe(b,a));
}

arbre_p :: arbre_p(int _n,int r,graphe_p *graph,arbre_p *arbre_ini)
{
	int a,k;
	
	n =  _n;
	root = r;
	
	if (!((pt = (int *)calloc(n + 1,sizeof(int))) && 
	(poidt = (int *)calloc(n + 1,sizeof(int))) &&
	(senst = (char *)calloc(n + 1,sizeof(char))) &&
	(dt = (int *)calloc(n + 1,sizeof(int))) &&
	(st = (int *)calloc(n + 1,sizeof(int)))))
		erreur(E_MEMOIRE);
		
	a = root;
	do
	{
		copie(a,pt,poidt,senst,dt,arbre_ini);
		
		k = arbre_ini->s(a);
		while((arbre_ini->d(k) > arbre_ini->d(root)) && (!arete(arbre_ini->p(k),k,graph)))
		{
			k = arbre_ini->s(arbre_ini->dernier(k));
		}
		
		if (arbre_ini->d(k) <= arbre_ini->d(root))
			st[a] = root;
		else
			st[a] = k;
		a = st[a];	
	}
	while(a != root);
}

int arbre_p :: taille()
{
	return(n);
}

int arbre_p :: racine()
{
	return(root);
}

int arbre_p :: p(int k)
{
	if ((k < 1) && (k > n))
		erreur(E_INDICE_NOEUD);
			
	return(pt[k]);
}

int arbre_p :: poid(int k)
{
	if ((k < 1) && (k > n))
		erreur(E_INDICE_NOEUD);
		
	return(poidt[k]);
}

char arbre_p :: sens(int k)
{
	if ((k < 1) && (k > n))
		erreur(E_INDICE_NOEUD);
		
	return(senst[k]);
}

int arbre_p :: d(int k)
{
	if ((k < 1) && (k > n))
		erreur(E_INDICE_NOEUD);
			
	return(dt[k]);
}

int arbre_p :: s(int k)
{
	if ((k < 1) && (k > n))
		erreur(E_INDICE_NOEUD);
			
	return(st[k]);
}

void arbre_p :: cycle(int ex,int ey,int *am,char *cotex,int *min,int *join)
{
	int ax,ay,dx,dy;

	ax = ex;
	ay = ey;
	*min = -1;
	do
	{
		dx = d(ax);
		dy = d(ay);

		if(dx >= dy)
		{
			if ((!sens(ax)) && ((poid(ax) <= *min) || (*min == -1)))
			{
				*am = ax;
				*min = poid(ax);
				*cotex = 1;
			}
			ax = p(ax);
		}
		if(dy >= dx)
		{
			if ((sens(ay)) && ((poid(ay) < *min) || (*min == -1)))
			{
				*am = ay;
				*min = poid(ay);
				*cotex = 0;
			}
			ay = p(ay);
		}
	}
	while(ax != ay);
	
	*join = ax;
}

void arbre_p :: poid_cycle(int ex,int ey,int min,int join)
{
	int a;

	a = ex;
	while(a != join)
	{
		if(sens(a))
			poidt[a] += min;
		else
			poidt[a] -= min;
		a = p(a);
	}

	a = ey;
	while(a != join)
	{
		if(sens(a))
			poidt[a] -= min;
		else
			poidt[a] += min;
		a = p(a);
	}
}

int arbre_p :: maj_s(int e1,int e2,int f1,int f2)
{
	int a,b,k,i,j,r;
	
	a = f1;
	while(s(a) != f2)
		a = s(a);
		
	b = s(e1);
	i = e2;
	
	k = i;
	while(d(s(k)) > d(i))
		k = s(k);
	r = s(k);
	
	while(i != f2)
	{
		j = i;
		i = p(i);
		st[k] = i;
		
		k = i;
		while(s(k) != j)
			k = s(k);
			
		if(d(r) > d(i))
		{
			st[k] = r;
			while(d(s(k)) > d(i))
				k = s(k);
			r = s(k);
		}
	}
	
	st[e1] = e2;
	if (e1 != a)
	{
		st[a] = r;
		st[k] = b;
	}
	else
		st[k] = r;
	
	return(k);
}

void arbre_p :: maj_p_poid_sens(int e1,int e2,int min,int se2,int f2)
{
	int k,pre,poidp,sensp,c,pc,poidc,sensc;
	
	pc = p(e2);
	poidc =  poid(e2);
	sensc = sens(e2);
	
	pt[e2] = e1;
	poidt[e2] = min;
	senst[e2] = se2;
	
	k = pc;
	
	pre = e2;
	poidp = poidc;
	sensp = sensc;
	
	while(pre != f2)
	{
		c = k;
		pc = p(k);
		poidc = poid(k);
		sensc = sens(k);
		
		pt[k] = pre;
		poidt[k] = poidp;
		senst[k] = !sensp;
		
		k = pc;
		
		pre = c;
		poidp =  poidc;
		sensp = sensc;
	}
}

void arbre_p :: maj_d(int e2,int dernier)
{
	int k = e2;
	
	while(k != dernier)
	{
		dt[k] = d(p(k)) + 1;
		k = s(k);
	}
	dt[dernier] = d(p(dernier)) + 1;
}
		
void arbre_p :: maj_arbre(int ex,int ey,int cotex,int min,int am)
{
	int e1,e2,dernier;
	
	if (cotex)
	{
		dernier = maj_s(ey,ex,p(am),am);
		maj_p_poid_sens(ey,ex,min,0,am);
		maj_d(ex,dernier);
	}
	else
	{
		dernier = maj_s(ex,ey,p(am),am);
		maj_p_poid_sens(ex,ey,min,1,am);
		maj_d(ey,dernier);
	}
}	 

void arbre_p :: entre(int ex,int ey,int *sx,int *sy,int *min)
{
	int *am,*join;
	char *cotex;
	int k;

	if (!((am = (int *) calloc(1,sizeof(int))) &&
	(cotex = (char *) calloc(1,sizeof(char))) &&
	(join = (int *) calloc(1,sizeof(int)))))
		erreur(E_MEMOIRE);

	cycle(ex,ey,am,cotex,min,join);

	if (sens(*am))
	{
		*sy = *am;
		*sx = p(*am);
	}
	else
	{
		*sx = *am;
		*sy = p(*am);
	}
	
	poid_cycle(ex,ey,*min,*join);
	
	maj_arbre(ex,ey,*cotex,*min,*am);
}

int arbre_p :: rech_dernier(int dsr,int noeud)
{
	if (d(s(noeud)) <= dsr)
		return(noeud);
	else
		rech_dernier(dsr,s(noeud));
}

int arbre_p :: dernier(int sous_racine)
{
	int dsr;
	
	if ((sous_racine < 1) && (sous_racine > n))
		erreur(E_INDICE_NOEUD);
	
	return(rech_dernier(d(sous_racine),sous_racine));
}
	
void arbre_p :: erreur(int n)
{
	switch(n)
	{
		case E_INDICE_NOEUD: printf("Erreur : indice de noeud incorrecte\n"); break;
		case E_N: printf("Erreur : valeur de n incorrecte\n"); break;
		case E_MEMOIRE: printf("Erreur : memoire pour un arbre\n");break;
	}
	exit(0);
}
