#include <stdio.h>
#include "arbre_p.H"

#define E_N_TI 1
#define E_COMPATIBLE 2 


class trans_ini
{
	private:
	int n;
	graphe_p *graph;
	arbre_p *arbre;
	
	public:
	trans_ini(int _n,graphe_p *_graph,arbre_p *_arbre);
	
	arbre_p *donne_arbre();
	
	private:
	void affiche(long int *c);
	void ecrit(int a);
	void affiche_arbre();
	void cree_c(long int *c);
	char trouve_e(int *ex,int *ey,int *gain,long int *c);
	
	public:
	int cout_arbre();
	void resolution();
	
	private:
	void erreur(int k);
};

trans_ini :: trans_ini(int _n,graphe_p *_graph,arbre_p *_arbre)
{
	if(n < 0)
		erreur(E_N_TI);
		
	if(_graph->taille() != _arbre->taille())
		erreur(E_COMPATIBLE);

	n = _n;
	graph = _graph;
	arbre = _arbre;
}

arbre_p *trans_ini :: donne_arbre()
{
	return(arbre);
}

int trans_ini :: cout_arbre()
{
	int a,r,res = 0;
	
	r = arbre->racine();
	a = arbre->s(r);
	
	while(a != r)
	{
		if (arbre->sens(a))
			res += arbre->poid(a) * graph->cout(arbre->p(a),a);
		else
			res += arbre->poid(a) * graph->cout(a,arbre->p(a));
		a = arbre->s(a);
	}
	
	return(res);
}

void trans_ini :: affiche(long int *c)
{
	int a;
	
	a = arbre->racine();
	
	do
	{
		printf("c[%d] = %d\n",a,c[a]);
		a = arbre->s(a);
	}
	while(a != arbre->racine());
	printf("\n");
}

void trans_ini :: ecrit(int a)
{
	if (a >= 10000)
		printf("%d",a);
	else
		if (a >= 1000)
			printf(" %d",a);
		else
			if (a >= 100)
				printf("  %d",a);
			else
				if (a >= 10)
					printf("   %d",a);
				else
					printf("    %d",a);
}

void trans_ini :: affiche_arbre()
{
	int a;
	
	printf("Description de l'arbre :\n");
	
	printf("   i");
	a = arbre->racine();
	do
	{
		ecrit(a);
		a = arbre->s(a);
	}
	while(a != arbre->racine());
	
	printf("\n   p");
	a = arbre->racine();
	do
	{
		ecrit(arbre->p(a));
		a = arbre->s(a);
	}
	while(a != arbre->racine());
	
	printf("\npoid");
	a = arbre->racine();
	do
	{
		ecrit(arbre->poid(a));
		a = arbre->s(a);
	}
	while(a != arbre->racine());
	
	printf("\nsens");
	a = arbre->racine();
	do
	{
		ecrit(arbre->sens(a));
		a = arbre->s(a);
	}
	while(a != arbre->racine());
	
	printf("\n   d");
	a = arbre->racine();
	do
	{
		ecrit(arbre->d(a));
		a = arbre->s(a);
	}
	while(a != arbre->racine());
	
	printf("\n   s");
	a = arbre->racine();
	do
	{
		ecrit(arbre->s(a));
		a = arbre->s(a);
	}
	while(a != arbre->racine());
	printf("\n\n");
}

void trans_ini :: cree_c(long int *c)
{
	int a,r;

	r = arbre->racine();
	c[r] = 0;

	a = arbre->s(r);
	while(a != r)
	{
		if (arbre->sens(a))
			c[a] = c[arbre->p(a)] + graph->cout(arbre->p(a),a);
		else
			c[a] = c[arbre->p(a)] - graph->cout(a,arbre->p(a));
		a = arbre->s(a);
	}
}

char trans_ini :: trouve_e(int *ex,int *ey,int *gain,long int *c)
{
	int x,y;
	
	x = arbre->racine();
	y = arbre->s(x);
	do
	{
		do
		{
			if (graph->existe(x,y) && (arbre->p(y) != x))
				if ((c[x] + (long int)graph->cout(x,y)) < c[y])
				{
					*ex = x;
					*ey = y;
					*gain = c[y] - c[x] - graph->cout(x,y);
					return(1);
				}
			if (graph->existe(y,x) && (arbre->p(y) != x))
				if ((c[y] + (long int)graph->cout(y,x)) < c[x])
				{
					*ex = y;
					*ey = x;
					*gain = c[x] - c[y] - graph->cout(y,x);
					return(1);
				}
			y = arbre->s(y);
		}
		while(y != arbre->racine());
		x = arbre->s(x);
		y = arbre->s(x);
	}
	while(arbre->s(x) != arbre->racine());
	
	return(0);
}

void trans_ini :: resolution()
{
	int r,k;
	long int *c;
	int *ex,*ey,*sx,*sy,*min,*gain;
	
	if (!(c = (long int *)calloc(n + 1,sizeof(long int))))
		erreur(E_MEMOIRE);
		
	affiche_arbre();
	printf("Le cout de cet arbre (solution) est : %d\n\n",cout_arbre());
/*	Initialisation de c	*/
	cree_c(c);
	affiche(c);
	printf("\n");
	
	if (!((ex = (int *) calloc(1,sizeof(int))) &&
	(ey = (int *) calloc(1,sizeof(int))) &&
	(gain = (int *) calloc(1,sizeof(int))) &&
	(sx = (int *) calloc(1,sizeof(int))) &&
	(sy = (int *) calloc(1,sizeof(int))) &&
	(min = (int *) calloc(1,sizeof(int)))))
		erreur(E_MEMOIRE);

/*	On realise la boucle tant qu'on trouve un arc entrant	*/ 		
	while(trouve_e(ex,ey,gain,c))
	{
		printf("On fait entrer l'arc (%d,%d)\n",*ex,*ey);
		printf("Le gain est de %d\n\n",*gain);
		
/*		Mise a jour de l'arbre	*/
		arbre->entre(*ex,*ey,sx,sy,min);
		
		printf("L'arc (%d,%d) de poid %d sort\n\n",*sx,*sy,*min);	
		affiche_arbre();
		printf("Le cout de cet arbre (solution) est : %d\n\n",cout_arbre());
		
/*		Mise a jour des couts	*/
		cree_c(c);	
		affiche(c);
		printf("\n");
	}
}

void trans_ini :: erreur(int k)
{
	switch(k)
	{
		case E_N_TI: printf("Erreur : n\n"); break;
		case E_COMPATIBLE: printf("Erreur : taille du graphe et taille de l'arbre incompatible\n"); break;
	}
	exit(0);
}
